REBOL [
	Title: "REBOL/Services - Async TCP Server Handler"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005, All rights reserved."
]

;-- Async Direct TCP Service Handler -----------------------------------------

add-server-handler [

	scheme: 'tcp

	opener: func [
		port [port!]
		config [object!]
	][
		debug connect ['opening-tcp-server port/port-id]
		port/user-data: config  ; server itself does not use a session here
		return open/direct/binary/async port :handler
	]

	handler: func [
		port action arg
		/local session
	][
		debug tcp ['handle-tcp-server action arg]
		session: port/user-data
		switch action [
			read [
				append session/recv-buf copy/part port arg
				debug tcp-read [arg 'bytes length? session/recv-buf 'total]
				if error? arg: try [
					while [ ; multiple requests may be concatenated in the buffer
						all [
							not empty? session/recv-buf
							arg: load-packet session
						]
					][
						if word? arg [
							send-fatal port arg
						]
						if block? arg [
							if not empty? session/changed [change-session session]
							arg: do-request session arg
							send-packet port arg 'reply
							if not empty? session/changed [change-session session]
						]
					]
				][
					send-fatal port reduce ['error disarm :arg]
				]
			]
			listen [
				arg: first port
				session: make-session arg port/user-data ;(config)
				session/remote-ip: arg/remote-ip
				debug connect ['open session/id arg/remote-ip arg/remote-port arg/local-port]
				if deny-ip? session/config arg/remote-ip [
					debug connect ['deny arg/remote-ip]
					send-fatal port "Access denied"
				]
			]
			close [
				debug connect ['close session/id port/remote-ip port/remote-port port/local-port]
				close-server port
			]
			error [
				attempt [send-fatal port reduce ['tcp-error disarm :arg]]
				close-server port
			]
		]
	]

	close-server: func [
		"Close server side of connection, free the session."
		port
	][
		attempt [close port]
		free-session port/user-data
	]

	send-fatal: func [
		"Low level packet failure response. (server only)"
		port [port!]
		error-info
	][
		debug error ['fatal error-info]
		send-packet port reduce ['fatal error-info]
		close port
	]

]